@@ -23,7 +23,6 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 
 #define NULL 0
 #define EXCEPTION_NAME_PREFIX "__exception_"
-#define EXCEPTION_NAME_LENGTH 12
 
 #include "config.h"
 #include "tree.h"
@@ -42,6 +41,9 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 #define _JBLEN (sizeof(jmp_buf)/sizeof(int))
 #endif
 
+void init_exception_processing ();
+void init_exception_processing_1 ();
+
 /* If non-zero, a VAR_DECL whose cleanup will cause a throw to the
    next exception handler.  Its value says whether to throw or not.
    In the case of functions which do not issue a RAISE, it should be
@@ -65,9 +67,9 @@ tree current_exception_object;
 /* Low-level rtl interface.  */
 #include "rtl.h"
 
-/* Cache `setjmp', `longjmp', and `unhandled_exception' after
-   default conversion.  Maybe later they will get built-in.  */
-static tree BISJ, BILJ, BIUE;
+/* Cache `setjmp', `longjmp', `raise_exception', and `unhandled_exception'
+   after default conversion.  Maybe later they will get built-in.  */
+static tree BISJ, BILJ, BIR, BIUE;
 
 /* Local variables which give the appearance that exception
    handling is part of the language and the execution model.  */
@@ -76,7 +78,7 @@ static tree BISJ, BILJ, BIUE;
 static tree EHS_type;
 
 /* The global handler stack.  */
-static tree EHS_decl;
+tree EHS_decl;
 
 /* Cached component refs to fields of `EHS_decl'.  */
 static tree EHS_prev, EHS_handler, EHS_parms, EHS_name;
@@ -137,7 +139,7 @@ lookup_exception_cname (ctype, cname, raise_id)
     this_cname = NULL_TREE;
   else if (TREE_CODE (this_cname) != IDENTIFIER_NODE)
     {
-      sorry ("multiple scope refs in `expand_cplus_raise_stmt'");
+      sorry ("multiple scope refs in `cplus_expand_raise_stmt'");
       this_cname = error_mark_node;
     }
   return this_cname;
@@ -402,10 +404,7 @@ finish_exception_decl (cname, decl)
     error ("cannot define exception inside function scope");
   else
     {
-      /* Now build the constructor for this exception.
-	 Pretending that CTOR is not a constructor simplifies
-	 things here.  */
-      DECL_CONSTRUCTOR_P (ctor) = 0;
+      /* Now build the constructor for this exception.  */
       parmdecls = DECL_ARGUMENTS (ctor);
       start_function (NULL_TREE, ctor, 0, 1);
       store_parm_decls ();
@@ -417,25 +416,19 @@ finish_exception_decl (cname, decl)
       /* Move all the parameters to the fields, skipping `this'.  */
       parmdecls = TREE_CHAIN (parmdecls);
       /* Install `name' of this exception handler.  */
-      expand_assignment (build (COMPONENT_REF, TREE_TYPE (fields), C_C_D, fields),
-			 build_unary_op (ADDR_EXPR, edecl, 0), 0, 0);
+      DECL_INITIAL (fields) = build_unary_op (ADDR_EXPR, edecl, 0);
       fields = TREE_CHAIN (fields);
       /* Install all the values.  */
       while (fields)
 	{
-	  expand_expr_stmt (build_modify_expr (build (COMPONENT_REF, TREE_TYPE (fields), C_C_D, fields),
-					       INIT_EXPR,
-					       parmdecls));
+	  /* Set up the initialization for this field.  */
+	  DECL_INITIAL (fields) = parmdecls;
 	  fields = TREE_CHAIN (fields);
 	  parmdecls = TREE_CHAIN (parmdecls);
 	}
-      c_expand_return (current_class_decl);
-      expand_end_bindings (getdecls (), 1, 0);
-      poplevel (1, 0, 1);
-      pop_momentary ();
+      emit_base_init (t, 0);
 
-      finish_function (DECL_SOURCE_LINE (ctor), 0);
-      DECL_CONSTRUCTOR_P (ctor) = 1;
+      finish_function (DECL_SOURCE_LINE (ctor), 1);
     }
 }
 
@@ -457,7 +450,9 @@ cplus_expand_start_try (implicit)
 
   /* Start a new block enclosing the whole handler.  */
   if (implicit)
-    pushlevel_temporary (1);
+    {
+      pushlevel_temporary (1);
+    }
   else
     {
       pushlevel (0);
@@ -476,6 +471,11 @@ cplus_expand_start_try (implicit)
   DECL_INITIAL (handler) = error_mark_node;
   finish_decl (handler, NULL_TREE, 0);
 
+  /* Must come after call to `finish_decl', else the cleanup for the temp
+     for the handler will cause the contour we just created to be popped.  */
+  if (implicit)
+    declare_implicit_exception ();
+
   /* Catch via `setjmp'.  */
   ref = build_component_ref (handler, get_identifier ("handler"), NULL_TREE, 0);
   call_to_setjmp = build_function_call (BISJ, build_tree_list (NULL_TREE, ref));
@@ -492,8 +492,7 @@ tree
 cplus_expand_end_try (keep)
      int keep;
 {
-  tree decls;
-  tree decl;
+  tree decls, decl, block;
 
   if (keep < 2)
     pop_implicit_try_blocks (NULL_TREE);
@@ -510,24 +509,30 @@ cplus_expand_end_try (keep)
     {
     case 0:
       expand_end_bindings (decls, 0, 1);
-      poplevel (0, 0, 0);
+      block = poplevel (0, 0, 0);
       pop_momentary (); 
       decl = getdecls ();
       break;
 
     case 1:
-      expand_end_bindings (decls, decls != 0, 1);
-      poplevel (decls != 0, 1, 0);
+      expand_end_bindings (decls, 1, 1);
+      block = poplevel (1, 1, 0);
       pop_momentary ();
       decl = getdecls ();
       break;
 
     default:
       decl = tree_last (decls);
+      block = NULL_TREE;
       break;
     }
 
   assert (TREE_CODE (decl) == VAR_DECL && TREE_TYPE (decl) == EHS_type);
+  if (block)
+    {
+      TREE_LANG_FLAG_1 (block) = 1;
+      TREE_USED (block) = 1;
+    }
 
   /* Pass it back so that its rtl can be bound to its name
      (or vice versa).  */
@@ -567,7 +572,7 @@ call_to_unhandled_exception ()
 {
   extern int lineno;
   tree parms = tree_cons (NULL_TREE,
-			  combine_strings (build_string (strlen (input_filename), input_filename)),
+			  combine_strings (build_string (strlen (input_filename + 1), input_filename)),
 			  build_tree_list (NULL_TREE, build_int_2 (lineno, 0)));
   return build_function_call (BIUE, parms);
 }
@@ -593,12 +598,11 @@ cplus_expand_end_except (dfault)
 	  expand_expr (call_to_unhandled_exception (), 0, VOIDmode, 0);
 	  expand_end_cond ();
 	}
+      /* Try the next handler.  */
+      if (! expand_escape_except ())
+	compiler_error ("except nesting botch");
     }
 
-  /* Try the next handler.  */
-  if (! expand_escape_except ())
-    compiler_error ("except nesting botch");
-
   raised = expand_end_except ();
 
   decls = getdecls ();
@@ -638,7 +642,7 @@ cplus_expand_end_except (dfault)
    If EXP is non-NULL, then it is an already constructed object
    of the kind that we want.  */
 void
-expand_cplus_raise (raise_id, parms, exp)
+cplus_expand_raise (raise_id, parms, exp)
      tree raise_id;
      tree parms;
      tree exp;
@@ -699,8 +703,17 @@ expand_cplus_raise (raise_id, parms, exp)
   /* Set the global exception handler stack's NAME field
      to the `name' of this exception.  The global exception
      handler stack is the container for the exception object
-     we just built.  */
+     we just built.
+
+     We go through a function call to make life easier when debugging.  */
+#if 0
   expand_assignment (EHS_name, build_unary_op (ADDR_EXPR, decl, 0), 0, 0);
+#else
+  parms = tree_cons (NULL_TREE, build_unary_op (ADDR_EXPR, EHS_name, 0),
+		     build_tree_list (NULL_TREE,
+				      build_unary_op (ADDR_EXPR, decl, 0)));
+  expand_expr (build_function_call (BIR, parms));
+#endif
 
   /* Activate thrower.  If we are inside a TRY statement,
      we can cheat and not do this, saving a longjmp.  */
@@ -722,8 +735,8 @@ expand_cplus_raise (raise_id, parms, exp)
      will take care of everybody.  */
 }
 
-void
-expand_cplus_start_catch (raise_id)
+tree
+cplus_expand_start_catch (raise_id)
      tree raise_id;
 {
   tree cname = lookup_exception_cname (current_class_type, current_class_name, raise_id);
@@ -731,7 +744,10 @@ expand_cplus_start_catch (raise_id)
   tree ref, cond;
 
   if (cname == error_mark_node)
-    cond = error_mark_node;
+    {
+      decl = error_mark_node;
+      cond = error_mark_node;
+    }
   else
     {
       decl = lookup_exception_object (cname, TREE_VALUE (raise_id), 1);
@@ -747,6 +763,7 @@ expand_cplus_start_catch (raise_id)
 	}
     }
   expand_start_cond (cond, 0);
+
   /* Does nothing right now.  */
   expand_catch (decl);
   if (current_exception_type
@@ -757,17 +774,18 @@ expand_cplus_start_catch (raise_id)
       expand_start_bindings (0);
       expand_decl_cleanup (NULL_TREE, cleanup);
     }
+  return decl;
 }
 
 void
-expand_cplus_end_catch (for_reraise)
+cplus_expand_end_catch (for_reraise)
      int for_reraise;
 {
   if (current_exception_type
       && TYPE_NEEDS_DESTRUCTOR (current_exception_type))
     {
       /* Destroy the specific exception object now in scope.  */
-      expand_end_bindings (getdecls (), 1, 1);
+      expand_end_bindings (getdecls (), 0, 1);
     }
   if (for_reraise)
     {
@@ -782,11 +800,33 @@ expand_cplus_end_catch (for_reraise)
   expand_end_cond ();
 }
 
+/* Reraise an exception.
+   If EXCEPTIONS is NULL_TREE, it means reraise whatever exception was caught.
+   If EXCEPTIONS is an IDENTIFIER_NODE, it means reraise the exception
+   object named by EXCEPTIONS.  This must be a variable declared in
+   an `except' clause.
+   If EXCEPTIONS is a TREE_LIST, it is the list of exceptions we are
+   willing to reraise.  */
+
 void
-expand_cplus_reraise (exceptions)
+cplus_expand_reraise (exceptions)
      tree exceptions;
 {
-  tree current_exception_ptr;
+  tree ex_ptr;
+  tree ex_object = current_exception_object;
+
+  if (exceptions && TREE_CODE (exceptions) == IDENTIFIER_NODE)
+    {
+      /* Don't get tripped up if its TREE_TYPE is `error_mark_node'.  */
+      ex_object = IDENTIFIER_LOCAL_VALUE (exceptions);
+      if (ex_object == NULL_TREE || TREE_CODE (ex_object) != INDIRECT_REF)
+	{
+	  error ("`%s' is not an exception decl", IDENTIFIER_POINTER (exceptions));
+	  return;
+	}
+      assert (TREE_CODE (TREE_OPERAND (ex_object, 0)) == VAR_DECL);
+      exceptions = NULL_TREE;
+    }
 
   /* reraise ALL, used by compiler.  */
   if (exceptions == NULL_TREE)
@@ -806,16 +846,14 @@ expand_cplus_reraise (exceptions)
 	  expand_assignment (exception_throw_decl, integer_one_node, 0, 0);
 	}
       /* Set to zero so that destructor will not be called.  */
-      expand_assignment (build (NOP_EXPR, ptr_type_node,
-				TREE_OPERAND (current_exception_object, 0)),
+      expand_assignment (build1 (NOP_EXPR, ptr_type_node, TREE_OPERAND (ex_object, 0)),
 			 integer_zero_node, 0, 0);
       if (! expand_escape_except ())
 	abort ();
       return;
     }
 
-  current_exception_ptr = build (NOP_EXPR, NULL_TREE,
-				 TREE_OPERAND (current_exception_object, 0));
+  ex_ptr = build1 (NOP_EXPR, NULL_TREE, TREE_OPERAND (ex_object, 0));
 
   /* reraise from a list of exceptions.  */
   while (exceptions)
@@ -827,24 +865,24 @@ expand_cplus_reraise (exceptions)
 	  error ("`%s' is not an exception type",
 		 IDENTIFIER_POINTER (TREE_VALUE (exceptions)));
 	  current_exception_type = NULL_TREE;
-	  TREE_TYPE (current_exception_object) = error_mark_node;
-	  TREE_TYPE (current_exception_ptr) = error_mark_node;
+	  TREE_TYPE (ex_object) = error_mark_node;
+	  TREE_TYPE (ex_ptr) = error_mark_node;
 	}
       else
 	{
 	  current_exception_type = type;
 	  /* In-place union.  */
-	  TREE_TYPE (current_exception_object) = type;
-	  TREE_TYPE (current_exception_ptr) = TYPE_POINTER_TO (type);
+	  TREE_TYPE (ex_object) = type;
+	  TREE_TYPE (ex_ptr) = TYPE_POINTER_TO (type);
 	}
 
       /* Now treat reraise like catch/raise.  */
-      expand_cplus_start_catch (exceptions);
-      expand_cplus_raise (exceptions, NULL_TREE, current_exception_ptr);
+      cplus_expand_start_catch (exceptions);
+      cplus_expand_raise (exceptions, NULL_TREE, ex_ptr);
       /* Set to zero so that destructor will not be called.  */
-      if (TREE_TYPE (current_exception_ptr) != error_mark_node)
-	expand_assignment (current_exception_ptr, integer_zero_node, 0, 0);
-      expand_cplus_end_catch (1);
+      if (TREE_TYPE (ex_ptr) != error_mark_node)
+	expand_assignment (ex_ptr, integer_zero_node, 0, 0);
+      cplus_expand_end_catch (1);
       exceptions = TREE_CHAIN (exceptions);
     }
   /* Don't propagate any unhandled exceptions.  */
@@ -881,13 +919,12 @@ init_exception_processing ()
   tree cname = get_identifier ("ExceptionHandler");
   tree field, chain;
   tree ctor, dtor;
-  tree EHS_DECL;
   tree jmp_buf_type = build_array_type (integer_type_node,
 					build_index_type (build_int_2 (_JBLEN-1, 0)));
   tree jmp_buf_arg_type = build_pointer_type (integer_type_node);
 
   tree parmtypes = hash_tree_chain (jmp_buf_arg_type, NULL_TREE);
-  tree setjmp_fndecl, longjmp_fndecl;
+  tree setjmp_fndecl, longjmp_fndecl, raise_fndecl;
 
   EHS_type = xref_tag (record_type_node, cname, NULL_TREE);
 
@@ -902,7 +939,12 @@ init_exception_processing ()
 							 hash_tree_chain (jmp_buf_arg_type,
 									  hash_tree_chain (integer_type_node, NULL_TREE))),
 				    NOT_BUILT_IN, 0);
+  raise_fndecl = define_function ("__raise_exception",
+				  build_function_type (void_type_node,
+						       hash_tree_chain (ptr_type_node,
+									hash_tree_chain (build_pointer_type (ptr_type_node), NULL_TREE))));
   BILJ = default_conversion (longjmp_fndecl);
+  BIR = default_conversion (raise_fndecl);
   BIUE = default_conversion (unhandled_exception_fndecl);
 
   pop_lang_context ();
@@ -943,18 +985,42 @@ init_exception_processing ()
   TYPE_HAS_CONSTRUCTOR (EHS_type) = 1;
   TYPE_HAS_DESTRUCTOR (EHS_type) = 1;
   finish_struct (EHS_type, temp_tree_cons (NULL_TREE, chain, NULL_TREE), 0, 0);
-  EHS_decl = build_decl (VAR_DECL, get_identifier ("exceptionHandlerStack"),
-			 TYPE_POINTER_TO (EHS_type));
-  /* If we don't push this, its definition, should it be encountered,
-     will not be seen.  */
-  EHS_decl = pushdecl (EHS_decl);
-  EHS_DECL = build1 (INDIRECT_REF, EHS_type, EHS_decl);
-  TREE_EXTERNAL (EHS_decl) = 1;
-  TREE_STATIC (EHS_decl) = 1;
-  TREE_PUBLIC (EHS_decl) = 1;
-  finish_decl (EHS_decl, 0, 0);
-  EHS_prev = build_component_ref (EHS_DECL, get_identifier ("prev"), 0, 0);
-  EHS_handler = build_component_ref (EHS_DECL, get_identifier ("handler"), 0, 0);
-  EHS_parms = build_component_ref (EHS_DECL, get_identifier ("parms"), 0, 0);
-  EHS_name = build_component_ref (EHS_DECL, get_identifier ("name"), 0, 0);
+}
+
+void
+init_exception_processing_1 ()
+{
+  register tree EHS_id = get_identifier ("exceptionHandlerStack");
+
+  EHS_decl = IDENTIFIER_GLOBAL_VALUE (EHS_id);
+
+  /* If we have no other definition, default to library implementation.  */
+  if (EHS_decl == NULL_TREE)
+    {
+      EHS_decl = build_decl (VAR_DECL, EHS_id, TYPE_POINTER_TO (EHS_type));
+      /* If we don't push this, its definition, should it be encountered,
+	 will not be seen.  */
+      EHS_decl = pushdecl (EHS_decl);
+      TREE_EXTERNAL (EHS_decl) = 1;
+      TREE_STATIC (EHS_decl) = 1;
+      TREE_PUBLIC (EHS_decl) = 1;
+      finish_decl (EHS_decl, 0, 0);
+    }
+  else if (TREE_CODE (EHS_decl) != VAR_DECL
+	   || TREE_TYPE (EHS_decl) != TYPE_POINTER_TO (EHS_type))
+    fatal ("exception handling declarations conflict with compiler's internal model");
+
+  if (EHS_prev == NULL_TREE)
+    {
+      register tree EHS_DECL = build1 (INDIRECT_REF, EHS_type, EHS_decl);
+      EHS_prev = build_component_ref (EHS_DECL, get_identifier ("prev"), 0, 0);
+      EHS_handler = build_component_ref (EHS_DECL, get_identifier ("handler"), 0, 0);
+      EHS_parms = build_component_ref (EHS_DECL, get_identifier ("parms"), 0, 0);
+      EHS_name = build_component_ref (EHS_DECL, get_identifier ("name"), 0, 0);
+    }
+
+#ifdef DBX_DEBUGGING_INFO
+  if (use_gdb_dbx_extensions)
+    dbxout_eh_init (EHS_type, EHS_decl);    
+#endif
 }
